'''
Widgets for managing lists of items. 
'''

import tkinter as tk
import tkinter.simpledialog
from tk_steroids.elements import Listbox, ButtonsFrame


class ItemManager(tk.Frame):
    '''
    Adding, removing, renaming items in a list.
    
    Attributes
    ----------
    current : string or None
        Name of the currently selected item (item_name)
    data : dict of dicts
        {item_name: {field1: string1, ...}}
    fields : list of strings
        Names of the fields user can input data (in addition to the
        name of the entry)
    listbox : object
        tk_steroids Listbox (tkinter Listbox object by self.listbox.listbox)
    buttons : object
        tk_steroids ButtonsFrame (self.buttons.buttons for underlying tkinter
        button objects)
    
    selection_callback : callable
        Called when an item is selected
    postchange_callback : callable
        Called after an item is removed, renamed or added
    save_callback : callable
        Callable when pressing save button
    cancel_callback : callable
        Callable when pressing close button
    '''

    def __init__(self, parent, fields=[], start_data={},
            selection_callback=None, postchange_callback=None,
            save_callback=None, cancel_callback=None):
        '''
        parent : object
            Tkinter parent widget
        start_data : dict of dicts OR dict of lists
            Initial entries in form {item_name: {field1: string1, ...}}
            or {item_name: [string1, string2, ...]}
        '''

        tk.Frame.__init__(self, parent)
        
        self.current = None
        self.fields = fields
        self.data = start_data
        self.selection_callback = selection_callback
        self.postchange_callback = postchange_callback
        
        # Left side entries

        self.listbox = Listbox(self, list(self.data.keys()), self._on_listbox_selection)
        self.listbox.grid(row=1, column=1, sticky='NSWE')

        
        # Right side buttons frame
        button_names = ['Add', 'Remove']
        button_commands = [self.add, self.remove]
        
        self.buttons = ButtonsFrame(self, button_names, button_commands, horizontal=False)
        self.buttons.grid(row=1, column=2)
        
        if callable(save_callback):
            tk.Button(self, text='Save', command=save_callback).grid(row=2, column=1)
        if callable(cancel_callback):
            tk.Button(self, text='Cancel', command=cancel_callback).grid(row=2, column=2)
        
        # Specify stretching
        self.columnconfigure(1, weight=1)
        self.rowconfigure(1, weight=1)


    def add(self):
        name = tkinter.simpledialog.askstring('New entry', 'Item name')
        if name:
            self.data[name] = {}

        self.listbox.set_selections(list(self.data.keys()))
        
        if self.postchange_callback:
            self.postchange_callback(self.data)
    

    def remove(self):
        
        del self.data[self.current]
        
        self.listbox.set_selections(list(self.data.keys()))
        
        if self.postchange_callback:
            self.postchange_callback(self.data)
 

    def _on_listbox_selection(self, selection):
        self.current = selection
        
        if self.selection_callback:
            self.selection_callback(self.data[selection])
   

    def set_data(self, data):
        '''
        data : list or dict
            [item1, item2, ...] or {item1: {}, item2: {}}
            where item_i is a string.
        '''
        if isinstance(data, list):
            self.data = {key: {} for key in data}
        else:
            self.data = data
        
        self.listbox.set_selections(list(self.data.keys()))
   


class ListManager(tk.Frame):
    '''
    Like ItemManager but each item is a further sublist.
    See ItemManager for documentation.
    '''
    
    def __init__(self, parent, start_data=[],
            save_callback=None, cancel_callback=None):
        tk.Frame.__init__(self, parent)

        self.im2 = ItemManager(self,
                save_callback=save_callback, cancel_callback=cancel_callback)
        self.im1 = ItemManager(self, start_data=start_data,
                selection_callback=self.im2.set_data)

        self.im2.postchange_callback = self.update_im1

        self.im1.grid(row=1, column=1, sticky='NSWE')
        self.im2.grid(row=1, column=2, sticky='NSWE')

        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)
        self.rowconfigure(1, weight=1)


    def update_im1(self, data):
        self.im1.data[self.im1.current] = list(data.keys())
        self.im1.listbox.set_selections(list(self.im1.data.keys()))





import sys
import io

import tkinter as tk
from tk_steroids.elements import BufferShower

oldout = sys.stdout
sys.stdout = io.StringIO()
print('test')
print('test2')
sys.stdout.write('aana')

root = tk.Tk()
buf = io.StringIO()
BufferShower(root, sys.stdout).grid()
root.mainloop()



import tkinter as tk


def popup(tk_parent, widget_class, args=[], kwargs={}, title=None,
        grid_pos=(1,1), sticky='NSWE', grid_weights=(1,1),
        close_button=True):
    '''
    Launch a (uninitialized) tkinter widget in a Toplevel window,
    using grid positioning system.


    Options
    -------
    tk_parent : object
        Parent widget
    widget_class: class
        The (uninitialized) widget to be opened in the toplevel
    args: list
        Additional arguments passed to the widget, corresponding to
        widget(toplevel, args[0], args[1], ...)
    kwargs : dict
        Additional keyword arguments to the widget.
    title : string
        Optional title to the widow
    grid_pos : tuple
        Position (row, column) of the created widget
    sticky : string
        Tkinter grid sticky option for the widget
    grid_weights: tuple
        Row and column weights for grid_{row,column}configure at grid_pos
    close_button : bool
        If True, add close button

    Returns
    -------
    toplevel: object
        The toplevel window
    widget : object
        The initialized widget
    '''
    toplevel = tk.Toplevel(tk_parent)
    if title:
        toplevel.title(title)
    
    toplevel.grid_rowconfigure(grid_pos[0], weight=grid_weights[0])
    toplevel.grid_columnconfigure(grid_pos[1], weight=grid_weights[1])
    
    widget = widget_class(toplevel, *args, **kwargs)
    widget.grid(row=grid_pos[0], column=grid_pos[1], sticky=sticky)
    
    tk.Button(toplevel, text='Close', command=toplevel.destroy).grid(row=2, column=1)
    
    return toplevel, widget


def popup_tickselect(tk_parent, *args, title='Make selection', **kwargs):
    top = tk.Toplevel(tk_parent)
    top.title(title)
    top.grid_columnconfigure(0, weight=1)
    top.grid_rowconfigure(1, weight=1)
 
    selector = TickSelect(top, *args, **kwargs)
    selector.grid(sticky='NSEW')

    tk.Button(selector, text='Close', command=top.destroy).grid(row=2, column=1)



class TickSelect(tk.Frame):
    '''
    User sets ticks to select items from selections group and
    presses ok -> callback_on_ok gets called as the made selections list
    as the only input argument.
    '''

    def __init__(self, parent, selections, callback_on_ok, close_on_ok=True, ticked=None,
            callback_args=[], callback_kwargs={}, single_select=False, search=10):
        '''
        selections          List of strings
        callback_on_ok      Callable, whom a sublist of selections is passed
        close_on_ok         Call root.destroy() when pressing ok
        ticked              A sublist of selections that should be enabled by default.
        callback_args       A list of secondary callback arguments, passed after the selections
        callback_kwargs     A dict of callback keyword arguments
        single_select       Use Radiobuttons instead Checkbutton, allowing only one item to be selected.
        search : int or bool
            If True, show the seach bar, false hide. If integer, specifies the number
            of selections when to start showing the search bar.
        '''
        tk.Frame.__init__(self, parent)

        self.grid_rowconfigure(0, weight=1) 
        self.grid_columnconfigure(0, weight=1) 
        
        self.callback_on_ok = callback_on_ok
        self.selections = selections
        self.close_on_ok = close_on_ok

        self.callback_args = callback_args
        self.callback_kwargs = callback_kwargs

        # Add scrollbar - adds canvas and extra frame
        canvas = tk.Canvas(self)
        frame = tk.Frame(canvas)
        frame.grid(row=0, column=0, sticky='NSEW')

        scrollbar = tk.Scrollbar(self, orient='vertical', command=canvas.yview)
        scrollbar.grid(row=0, column=1, sticky='NS')
        
        canvas.configure(yscrollcommand=scrollbar.set)
        canvas.create_window((0,0), window=frame, anchor='nw')
        
        canvas.grid_rowconfigure(0, weight=1)
        canvas.grid(row=0, column=0, sticky='NSEW')
        
        self.checkbuttons = []

        # Create tickboxes and entries
        if single_select:
            N_selections = 1
        else:
            N_selections = len(self.selections)
        tk_variables = [tk.IntVar() for i in range(N_selections)]

        for i_row, selection in enumerate(self.selections):        
            if single_select:
                checkbutton = tk.Radiobutton(frame, text=selection, variable=tk_variables[0], value=i_row)
            else:
                checkbutton = tk.Checkbutton(frame, text=selection, variable=tk_variables[i_row])
            checkbutton.grid(sticky='W')
            
            # Set ticked
            if not ticked is None and selection in ticked:
                checkbutton.select()

            self.checkbuttons.append(checkbutton)
    
        # Buttons under the selections list
        self.buttons_frame = tk.Frame(self)
        tk.Button(self, text='Ok', command=self.on_ok).grid(row=1, column=1)
        tk.Button(self.buttons_frame, text='Select all', command=self.select_all).grid(row=1, column=0)
        tk.Button(self.buttons_frame, text='Inverse', command=self.toggle_selection).grid(row=1, column=1)
        self.buttons_frame.grid(row=1, column=0)

        if (search is True) or (search is not False and search < len(self.selections)):
            self.last_search = ''
            self.searchtext = tk.StringVar()
            self.searchbox = tk.Entry(self, text='Search', textvariable=self.searchtext)
            self.searchbox.grid(row=2, column=0, sticky='WE')
            self.winfo_toplevel().after(2000, lambda: self._update_search(repeat_after=1000))

        self.winfo_toplevel().after(50, self._update)
        
        self.frame = frame
        self.canvas = canvas
        self.tk_variables = tk_variables
        self.single_select = single_select
        
        self.visible_checkbuttons = self.checkbuttons


    def _update_search(self, repeat_after=None):
        '''
        Check the searchbox input and hide selections that do not match
        the search.
        '''
        key = str(self.searchtext.get())
        if key != self.last_search:
            
            self.visible_checkbuttons  = []

            for checkbutton in self.checkbuttons:
                checkbutton.grid_remove()
                if key in checkbutton.cget('text'):
                    checkbutton.grid()

                    self.visible_checkbuttons.append(checkbutton)

            self.last_search = key

        if repeat_after:
            self.winfo_toplevel().after(repeat_after, lambda r=repeat_after: self._update_search(repeat_after=r))


    def _update(self):
        self.canvas.config(scrollregion=(0, 0, self.frame.winfo_reqwidth(), self.frame.winfo_reqheight()))
        self.winfo_toplevel().after(1000, self._update)


    def select_all(self):
        for checkbutton in self.visible_checkbuttons:
            checkbutton.select()


    def toggle_selection(self):
        for checkbutton in self.visible_checkbuttons:
            checkbutton.toggle()


    def on_ok(self):
        '''
        Gets called when the OK button is pressed, and calls callback_on_ok with
        the made selections.
        '''
        made_selections = []

        if self.single_select:
            made_selections.append(self.selections[self.tk_variables[0].get()])
        else:
            for tk_variable, selection in zip(self.tk_variables, self.selections):
                if tk_variable.get() == 1:
                    made_selections.append(selection)

        self.callback_on_ok(made_selections, *self.callback_args, **self.callback_kwargs)
        

        if self.close_on_ok:
            self.winfo_toplevel().destroy()


def main():
    pass

if __name__ == "__main__":
    main()



import unittest

import tkinter as tk

import tk_steroids.menumaker
from tk_steroids.menumaker import MenuMaker


class SimpleMenu(MenuMaker):
    '''
    A simple method addition to the MenuMaker should lead
    to a menu with item labelled "Return none".
    '''

    def return_none(self):
        return None



class TestMenuMaker(unittest.TestCase):


    def setUp(self):
        
        self.menu = tk.Menu()
        self.maker = SimpleMenu('testmenu')

    
    def test_connect_disconnect(self):

        self.maker._connect(self.menu)
        self.maker._disconnect()
        
        # Important to test reconnecting after disconnecting
        self.maker._connect(self.menu)
        self.maker._disconnect()

    
    def test_enable_disable(self):

        self.maker._connect(self.menu)
        
        self.maker._disable()
        self.maker._enable()

        self.maker._disconnect()

  





import inspect

import tkinter as tk


class MenuMaker:
    '''
    An alternative way to make an tkinter menu structures, with an aim to
    improve code readability.


    Intended usage

        1) Create a new class inheriting from MenuMaker class
        2) Add new methods. These become the menu items (see the NOTE1)
        3) Create an object and call _connect(tk_parent_menu)

        NOTE1: About method naming:
            - Method names starting with underscore are excluded from the menu
                (all the original methods start with an unserscore)
            - Menu items are named after the method names
            - Method "use_underscore_names" would become "Use underscore names"
            - Mofidy replacement_dict attribute if you want to change this

    
    Ordering menu entries
        The default order is the order as in the class methods were defined.
        
        This can be changed by chaning the order attribute (see below)

        To have completly custom ordering, set order to "force_order" and
            make _force_order method to return the menu entry names in the wanted order


    (Private) attributes
    ------------------
    self.name
        Name of the menu
    self.replacement_dict
        A dictionary. Keys are strings in method names to be replaced by
        the values. Default is {'_': ' '}
    self.tkmenu
        None when not connected, otherwise a tkinter.Menu object
    self.added_menu_items
        A list of menu item names added to the menu (labels, string)
        For separators, index of the separator when added
    order : string or callable
        "alphabetical", "as_defined" or "force_order" or callable.
        If force_order but force_order is not redefined to return anything
        useful, uses 'as_defined'.
        If callable, the raw menuentry names are passed to the callable
        and sorted names are expected to be returned.
    '''


    def __init__(self, name, order='as_defined'):
        '''
        name        Name of the menu as shown on screen.
        '''
        
        self.name = name
        self.order = order
        self.replacement_dict = {'_': ' '}

        self.parent_menu = None
        self.tkmenu = None
        self.added_menu_items = [] 
        

    def _connect(self, tk_parent_menu, **kwargs):
        '''
        Create a new menu, populate it with the commands, and
        connect the menu to the tkmenu.

        **kwargs are passed to the created menu.
        '''

        self.parent_menu = tk_parent_menu
        self.tkmenu = tk.Menu(tk_parent_menu, **kwargs)

        method_names = self._force_order()
                
        all_method_names = self._list_items(fancy_names=False)
        
        if method_names is None:
            # In case self._force_order not overridden
            method_names = all_method_names
        else:
            # Make sure to add also those methods that were
            # not speified in self._force_order
            method_names += [name for name in all_method_names if name not in method_names]


        for method_name in method_names:
            
            if method_name == '.':
                
                self.tkmenu.add_separator()

                index = self.tkmenu.index(tk.END)
                self.added_menu_items.append(index)

            else:
                method = getattr(self, method_name)
                fancyname = self._fancy_name(method_name)

                self.tkmenu.add_command(label=fancyname, command=method)
                
                self.added_menu_items.append(fancyname)
        
        self.parent_menu.add_cascade(label=self.name, menu=self.tkmenu)



    def _disconnect(self):
        '''
        Remove menu items added by this menumaker from the menu.
        '''
        
        if self.tkmenu is None:
            raise ValueError("MenuMaker not connected to a tkinter.Menu object; Cannot disconnect.")


        for item in self.added_menu_items:
            self.tkmenu.delete(item)

        self.parent_menu.delete(self.name)

        self.parent_menu = None
        self.tkmenu = None
        self.added_menu_items = []



    def _fancy_name(self, name):
        '''
        Make a method name into a fancy name.
        
        By default, underscores are replaced by spaces and the name
        is capitalized.
        '''
        fancyname = name[0].upper() + name[1:]

        for original, replacement in self.replacement_dict.items():
            fancyname = fancyname.replace(original, replacement)
        
        return fancyname



    def _list_items(self, fancy_names=False):
        '''
        Returns a list of menu item names (in alphabetical order).

        If fancy_names == True,
        returns the method names as they are shown in the menu.
        '''

        method_names = [method for method in dir(self)
                if not method.startswith('_') and callable(getattr(self, method))]
        
        if self.order == 'alphabetical':
            method_names.sort()
        elif self.order in ['as_defined', 'force_order']:
            method_names = self._sort_by_definition_order(method_names)
        elif callable(self.order):
            method_names = self.order(method_names)

        if fancy_names:
            method_names = [self._fancy_name(name) for name in method_names]

        return method_names
        


    def _enable(self):
        '''
        Enable (make clickable) the menu entries added by this menumaker.
        '''

        if self.tkmenu is None:
            raise ValueError("MenuMaker not connected to a tkinter.Menu object; Cannot enable items.")
        
        raise NotImplementedError('Enabling not implemented yet')



    def _disable(self):
        '''
        Disable (make unclickable) the menu entries added by this menumaker.
        '''
        if self.tkmenu is None:
            raise ValueError("MenuMaker not connected to a tkinter.Menu object; Cannot disable items.")
        
        raise NotImplementedError('Disabling not implemented yet')



    def _force_order(self):
        '''
        Override this method to force order by returning an interable of the
        methods (menu item) names (you can get the list of all names using _list_items)
        
        '.' represents a menu separator.

        For example, return ['open_window', 'close_some_window', '.', 'exit']
        to have separator in between "Close some window" and "Exit"

        '''
        return None


    def _sort_by_definition_order(self, method_names):
        '''
        Check the source code for definition order of the methods

        Can make mistakes if for example comments contain funtion definitions,
        but this merely leads into wrong ordering.

        Returns
        -------
        method_names : list of strings
            the method names in the defined order
        '''
        indices = {}
        
        names_to_look = [name for name in method_names]
        
        lines, linen = inspect.getsourcelines(self.__class__)

        for i_line, line in enumerate(lines):
            matches = [('def '+name in line) for name in names_to_look]
            
            if any(matches):
                index = matches.index(True)
                indices[names_to_look[index]] = i_line
                names_to_look.pop(index)
        
        return sorted(method_names, key=lambda x: indices.get(x, -1))





import tkinter as tk


CALLBACK_FORMATS = ['rgb', 'hex']

class _ColorMap(tk.Frame):
    '''
    A colorpicker part: Pick colors from a colormap.
    '''   

    def __init__(self, tk_parent, callback=None,
            callback_format='rgb'):
        '''
        callback : None or callable
            Callback on click on the colormap that takes in one positional
            argument that is the selected color.
        callback_format : string
            'rgb' or 'hex'
        '''
        tk.Frame.__init__(self, tk_parent)
        self.callback = callback

        if callback_format in ['rgb', 'hex']:
            self.callback_format = callback_format
        else:
            raise ValueError("Unkown callback format {}. Use {}".format(
                callback_format, CALLBACK_FORMATS))

        self.mouse_down = False

        self.map_width = 100
        self.map_height = 100
        self.colormap = tk.PhotoImage(height=self.map_height,
                width=self.map_width)
        
        red, green, blue = self._hsv()
        for j in range(0, self.map_height):
            saturation = 1-j/(self.map_height-1)
            for i in range(0, self.map_width):
                color = '#{0:02x}{1:02x}{2:02x}'.format(
                        int(saturation*red[i]+(1-saturation)*255),
                        int(saturation*green[i]+(1-saturation)*255),
                        int(saturation*blue[i]+(1-saturation)*255))
                self.colormap.put(color, to=(i,j))
        
        
        self.canvas = tk.Canvas(self, width=self.map_width, height=self.map_height)
        self.canvas.create_image(0, 0, image=self.colormap,
                anchor=tk.NW)
        self.canvas.grid(row=1, column=1)
            
        self.canvas.scale('all', 0, 0 ,4, 4)

        self.canvas.bind('<Button-1>', self._on_click)
       

    def _on_click(self, event):
        color = self.colormap.get(event.x, event.y)
        
        if self.callback_format == 'hex':
            color = '#{0:02x}{1:02x}{2:02x}'.format(*color)
        elif self.callback_format == 'rgb':
            pass

        if callable(self.callback):
            self.callback(color)
        

    def _hsv(self):
        '''
        Crete hsv colors.

        Returns three lists of intensity values (0-255)
        for the red, green and blue channels.
        '''
        # _   _
        #  \_/
        hsv_red = []
        for h_i in range(0, self.map_width):
            h = h_i * 360 / self.map_width
            if 0<=h< 60 or 300<=h<360: 
                value = 1
            elif 120<=h<240:
                value = 0
            elif 60<=h<120:
                value = (60-(h-60))/60
            elif 240<=h< 300:
                value = (h-240)/60

            hsv_red.append(int(255*value))
        
        # Blue
        h_cut = int(120/360*len(hsv_red))
        hsv_blue = hsv_red[h_cut:] + hsv_red[:h_cut]
        
        # Green
        h_cut = int(240/360*len(hsv_red))
        hsv_green = hsv_red[h_cut:] + hsv_red[:h_cut]
 
        return hsv_red, hsv_green, hsv_blue


class ColorPicker(tk.Frame):
    '''
    Pick colors from a colormap, show the picked color,
    and call callback on exit.
    
    Colors are common 8-bit 3-channel.
    '''

    def __init__(self, tk_parent, callback=None):
        tk.Frame.__init__(self, tk_parent)

        self.colormap = _ColorMap(self, callback=self.set_color,
                callback_format='hex')
        self.colormap.grid(row=1, column=1)
        
        self.preview = tk.Canvas(self, width=100, height=100)
        self.preview.grid(row=1, column=2)

    def set_color(self, color):
        '''
        Sets the currently selected color
        '''
        self.preview.config(bg=color)


class MultiPicker(tk.Frame):
    '''
    Displays the selected colors, and by pressing
    one of the colors, opens a ColorPicker instance.
    '''
    def __init__(self, tk_parent, color_names,
            callback=None):
        '''
        color_n
        callback : None or callable or list of callables
            Called when a color is set. If list
        '''
        tk.Frame.__init__(self, tk_parent)
        
        #for 
        #tk.Label(self, text=)

    





import tkinter as tk
import PIL
import PIL.Image
import PIL.ImageDraw
import PIL.ImageFont
import PIL.ImageTk

class ImageFeed(tk.Frame):
    '''
    A widgets that easily allows showing a camerafeed (or any PIL
    image feed) if the object has a method callled get, which returns
    a PIL image.
    '''

    def __init__(self, tk_master, size='original', feed_object=None, fallback_size=(800,600)):
        '''
        feed_object     Any object having get method that returns a PIL image
        size            Tuple of (x, y) in pixels or "original" or a float scaling
                        factor where values larger than 1 incease feed's size.
        '''
        tk.Frame.__init__(self, tk_master)
        self.tk_master = tk_master

        self.feed_object = feed_object
        
        if type(size) == type(('tuple',2)) and len(size) == 2:
            self.size = size
        elif type(size) == type(4.2) or type(size) == type(42):
            try:
                w, h = self.feed_object.get().size
            except:
                w, h = fallback_size
            self.size = (int(w*size), int(h*size))
        elif size == 'original':
            try:
                w, h = self.feed_object.get().size
            except:
                w, h = fallback_size
            self.size = (w, h)
        else:
            raise ValueError('Given size invalid: {}'.format(size))
    
        self.update_interval = 0

        self.imagelabel = tk.Label(self)
        self.imagelabel.grid()

    def set_source(self, feed_object):
        '''
        See __init__ method
        '''
        self.feed_object = feed_object

    def set_update_interval(self, milliseconds):
        '''
        If milliseconds. If 0 then no udpate.
        '''
        self.update_interval = milliseconds
        if self.update_interval:
            self.update_feed()

    def update_feed(self, img=None):
        if img is None:
            try:
                image = self.feed_object.get()
                image = image.resize(self.size)
                self.photoimage = PIL.ImageTk.PhotoImage(image)
            except Exception as e:
                image = PIL.Image.new('RGB', self.size)
                draw = PIL.ImageDraw.Draw(image)
                font = PIL.ImageFont.load_default()
                draw.text((0,0), "Error while retriving the image\n{}".format(str(e)), (255,255,255), font=font)
                
                self.photoimage = PIL.ImageTk.PhotoImage(image)
        else:
            self.photoimage = PIL.ImageTk.PhotoImage(img)

        self.imagelabel.configure(image=self.photoimage)
        
        if self.update_interval:
            self.after(self.update_interval, self.update_feed)

    def show_image(image):
        '''
        Instead of getting the image from the source, show any
        PIL object image.


        image       A PIL (pillow) image
        '''
        return self.update_feed(img=image)



'''
Connecting matplotlib to plot on tkinter canvases.
'''

import numpy as np

import tkinter as tk
import matplotlib.widgets
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import (
        RectangleSelector,
        PolygonSelector,
        EllipseSelector
        )
import matplotlib.ticker
from mpl_toolkits.mplot3d import proj3d

import collections

import matplotlib.patches


class ArrowSelector:
    '''
    Idea similar to matplotlib's RectangleSelector.
    '''

    def __init__(self, ax, callback, auto_connect=True):
        '''
        ax              Matlab axes intance
        callback        Gets eclick erelease, as in RectangleSelector
        auto_connect    Calls self.connect at initialization
        '''

        self.ax = ax
        self.callback = callback

        self.fig = ax.figure

        self.p0 = None
        self.p1 = None

        self.cid1 = None
        self.cid2 = None
        self.cid3 = None

        self.arrows = []
        
        if auto_connect:
            self.connect()



    def connect(self):
        '''
        Add ArrowSelector to the figure.

        Calls mpl_connect on self.fig.canvas 
        '''
        self.cid1 = self.fig.canvas.mpl_connect('button_press_event', self._on_press)
        self.cid2 = self.fig.canvas.mpl_connect('button_release_event', self._on_release)



    def disconnect(self):
        '''
        Does the opposite of connect, removes the rectangle selector
        from the figure.
        '''
        self.fig.canvas.mpl_disconnect(self.cid1)
        self.fig.canvas.mpl_disconnect(self.cid2)



    def _on_press(self, event):
        '''
        Called when pressing the figure with mouse.
        '''
        if event.inaxes == self.ax:
            self.p0 = (event.xdata, event.ydata)
            print(self.p0)

            # Set up updating the arrow
            self.cid3 = self.fig.canvas.mpl_connect('motion_notify_event', self._update_arrow)

        else:
            self.p0 = None



    def _clear_arrows(self):
        '''
        Remove any existing arrows from the figure
        '''
        if len(self.arrows) > 0:
            for arrow in self.arrows:
                arrow.remove()
            self.arrows = []
       


    def _update_arrow(self, event):
        '''
        Called when dragging the arrow around.
        '''
        if event.inaxes == self.ax:
            
            self._clear_arrows()
            
            # Arrow thickness
            width = 10

            dx = event.xdata - self.p0[0]
            dy = event.ydata - self.p0[1]
            arrow = matplotlib.patches.Arrow(*self.p0, dx, dy, width=width)
            
            # Add arrow to the figure
            self.ax.add_patch(arrow)
            self.arrows.append(arrow)

            self.fig.canvas.draw_idle()
        else:
            pass



    def _on_release(self, event):
        '''
        Called then releasing the mouse press.
        '''

        if self.cid3 is not None:
            self.fig.canvas.mpl_disconnect(self.cid3)
            self.cid3 = None

        if event.inaxes == self.ax and self.p0 is not None:
            self.p1 = (event.xdata, event.ydata)
            print(self.p1)
            
            # To roughly match RectangleSelector behaviour
            p0 = collections.namedtuple('eclick', ['xdata', 'ydata'])(self.p0[0], self.p0[1])
            p1 = collections.namedtuple('eclick', ['xdata', 'ydata'])(self.p1[0], self.p1[1])
            
            self._clear_arrows()

            self.callback(p0, p1)
        else:
            self.p0 = None
            self.p1 = None




class CanvasPlotter(tk.Frame):
    '''
    Embeds a matplotlib figure on a tkinter GUI.

    Attributes
    ----------
    self.figure : object
        Matplotlib Figure object
    self.ax : object
        Matplotlib Axes object
    self.parent : object
        Tkinter parent widget
    self.visibility_button : object
        A tkinter.Buttton to toggle show/hide
    '''
    
    def __init__(self, parent, text='', show=True, visibility_button=False,
            figsize=None,
            **kwargs):
        '''
        Creates a matplotlib figure and an axes objects when created, and then a
        FigureCanvasTkAgg

        ARGUMENTS
        ---------
        parent : object
            Tkinter parent widget
        text : string
            Title of the plot, to be shown in a Label Frame wrapping the plot
        show : bool
            
        projection : string
            See projection keyword argument for matplotlib's Figure.add_subplot
        '''

        tk.Frame.__init__(self, parent)
        self.parent = parent
        
        if figsize:
            self.figure = Figure(figsize=figsize)
        else:
            self.figure = Figure()
        self.ax = self.figure.add_subplot(111, **kwargs)
        
        self.frame = tk.LabelFrame(self, text=text)
        self.frame.grid(sticky='NSWE')
        
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.frame.grid_rowconfigure(1, weight=1)
        self.frame.grid_columnconfigure(0, weight=1)



        self.visibility_button = tk.Button(self.frame, text='', command=self.toggle_visibility)
        
        if visibility_button:
            self.visibility_button.grid(row=0, column=0, sticky='W')

        self.canvas = FigureCanvasTkAgg(self.figure, master=self.frame)
        #self.canvas.get_tk_widget().grid(sticky='NEWS') 
        self.canvas.draw()
        self.show()

        self.roi_callback = None
        self._previous_shape = None
        self._previous_roi_drawtype = None


    def get_figax(self):
        '''
        Returns the figure and ax so that plotting can be done externally.
        Remember to call update method!
        '''
        return self.figure, self.ax


    def plot(self, *args, ax_clear=True, **kwargs):
        '''
        For very simple plotting.

        Arguments
        ---------
        *args, **kwargs
            Directly passed to matplotlib plot method
        ax_clear : bool
            If True, clear the previous plottings away
        '''
        if ax_clear:
            self.ax.clear()
        self.ax.plot(*args, **kwargs)
        
        self.canvas.draw()


    def __onSelectRectangle(self, eclick, erelease):
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        self.roi_callback(x1, y1, x2, y2)
       
    def __onSelectPolygon(self, vertices):
        self.roi_callback(vertices)
        # FIXME
        # requires manual ESC press to start the selection process again

    
    def imshow(self, image, slider=False, normalize=True,
            roi_callback=None, roi_drawtype='box',
            **kwargs):
        '''
        Showing an image on the canvas, and optional sliders for colour adjustments.
        
        Redrawing image afterwards is quite fast because set_data is used
        instead imshow (matplotlib).

        INPUT ARGUMENTS
        slider          Whether to draw the sliders for setting image cap values
        roi_callback    A callable taking in x1,y1,x2,y2
        roi_drawtype    "box", "ellipse" "line" or "polygon"
        *kwargs     go to imshow

        Returns the object returned by matplotlib's axes.imshow.
        '''

        if image is None:
            image = self.imshow_image

        self.imshow_image = image
        
        # Slider
        if slider:
            # Check if the sliders exist. If not, create
            try:
                self.imshow_sliders
            except AttributeError:
                self.slider_axes = [self.figure.add_axes(rect) for rect in ([0.2, 0.05, 0.6, 0.05], [0.2, 0, 0.6, 0.05])]
                
                self.imshow_sliders = []
                self.imshow_sliders.append( matplotlib.widgets.Slider(self.slider_axes[0], 'Upper %', 0, 100, valinit=90, valstep=1) )
                self.imshow_sliders.append( matplotlib.widgets.Slider(self.slider_axes[1], 'Lower %', 0, 100, valinit=5, valstep=1) )
                for slider in self.imshow_sliders:
                    slider.on_changed(lambda slider_val: self.imshow(None, slider=slider, **kwargs))
            
            for ax in self.slider_axes:
                if ax.get_visible() == False:
                    ax.set_visible(True)
           
           
            # Normalize using the known clipping values
            #image = image - lower_clip
            #image = image / upper_clip
        else:
            # Hide the sliders if they exist
            if getattr(self, 'imshow_sliders', None):
                for ax in self.slider_axes:
                    if ax.get_visible() == True:
                        ax.set_visible(False)
                        print('axes not visible not')

        if getattr(self, 'imshow_sliders', None):
            # Check that the lower slider cannot go above the upper.
            if self.imshow_sliders[0].val < self.imshow_sliders[1].val:
                self.imshow_sliders[0].val = self.imshow_sliders[1].val

            upper_clip = np.percentile(image, self.imshow_sliders[0].val)
            lower_clip = np.percentile(image, self.imshow_sliders[1].val)
            image = np.clip(image, lower_clip, upper_clip)
 

        if normalize:
            image = image - np.min(image)
            image = image / np.max(image)


        # Just set the data or make an imshow plot
        if self._previous_shape == image.shape and (
                roi_callback is None or roi_drawtype == self._previous_roi_drawtype):
            self.imshow_obj.set_data(image)
        else:
            if hasattr(self, 'imshow_obj'):
                # Fixed here. Without removing the AxesImages object plotting
                # goes increacingly slow every time when visiting this else block
                # Not sure if this is the best fix (does it free all memory) but
                # it seems to work well
                self.imshow_obj.remove()

            self.imshow_obj = self.ax.imshow(image, **kwargs)
            self.figure.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
            self.ax.xaxis.set_major_locator(matplotlib.ticker.NullLocator()) 
            self.ax.yaxis.set_major_locator(matplotlib.ticker.NullLocator())
            if callable(roi_callback):

                if getattr(self, "roi_rectangle", None):
                    if self._previous_roi_drawtype == 'line':
                        self.roi_rectangle.disconnect()
                    else:
                        self.roi_rectangle.disconnect_events()

                if roi_drawtype == 'box':
                    self.roi_rectangle = RectangleSelector(self.ax, self.__onSelectRectangle,
                        useblit=True)
                elif roi_drawtype == 'ellipse':
                    self.roi_rectangle = EllipseSelector(self.ax, self.__onSelectRectangle,
                            useblit=True)
                elif roi_drawtype == 'line':
                    self.roi_rectangle = ArrowSelector(self.ax, self.__onSelectRectangle)
                elif roi_drawtype == 'polygon':
                    self.roi_rectangle = PolygonSelector(self.ax, self.__onSelectPolygon,
                            useblit=True)
                else:
                    raise ValueError('roi_drawtype either "box", "ellipse", "line", or "polygon", got {}'.format(roi_drawtype))
                
                self.roi_callback = roi_callback
                self._previous_roi_drawtype = roi_drawtype

        self._previous_shape = image.shape
        self.canvas.draw()
        
        return self.imshow_obj
    

    def hide(self):
        '''
        Hide the canvas widget.
        '''
        self.canvas.get_tk_widget().grid_forget()
        self.visible = False
        self.visibility_button.config(text='Show')


    def show(self):
        '''
        Show the cavas widget.
        (Not to be confused with matplotlib's show)
        '''
        self.canvas.get_tk_widget().grid(row=1, column=0, sticky='NSWE')
        self.visible = True
        self.visibility_button.config(text='Hide')
    

    def update(self):
        '''
        Call if any changes has made to the axes.
        '''
        self.canvas.draw()

    def update_size(self):
        '''
        Sets the frame size to match the matplotlib.Figure size.
        '''
        #self.canvas.config(width=800, height=400)
        w, h = self.figure.get_size_inches() * self.figure.dpi
        self.canvas.get_tk_widget().config(width=h, height=w)

        self.canvas.get_tk_widget().grid(row=1, column=0, sticky='NSWE')

    def toggle_visibility(self):
        '''
        Toggle wheter the plot is shown or hidden.
        '''

        if self.visible:
            self.hide()
        else:
            self.show()


class SequenceImshow(tk.Frame):
    '''
    Higher level CanvasPlotter.imshow for many images with a slider
    to select the current image to show.
    
    Attributes
    ----------
    parent : object
        Tkinter parent widget
    images : list of objects
        List of matplotlib imshow plottable objects.
    canvas_plotter : object
        CavasPlotter object
    slider : object
        Tkitner Scale (slider) for selecting the currently shown image.
    '''

    def __init__(self, parent, *args, **kwargs):
        
        tk.Frame.__init__(self, parent)
        
        self.images = []
        self.canvas_plotter = CanvasPlotter(self, *args, **kwargs)
        self.canvas_plotter.grid(row=1, column=1, sticky='NSWE')
        
        self.slider = tk.Scale(self, from_=1, to=1, orient=tk.HORIZONTAL,
                command=self.select_image)
        self.slider.grid(row=2, column=1, sticky='NSWE')
            
        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(1, weight=1)

        self._imshow_kwargs = {}


    def select_image(self, i_image=None):
        '''
        Callback for self.slider.
        '''
        if i_image is None:
            i_image = self.slider.get()

        self.canvas_plotter.imshow(self.images[int(i_image)-1],
                **self._imshow_kwargs)
        self.slider.set(i_image)


    def imshow(self, images, **kwargs):
        '''
        Set a new set of images to be shown and
        update the slider range.
        
        Arguments
        ---------
        kwargs : dict
            Keyword arguments for CanvasPlotter imshow or
            matplotlib's imshow.
        '''
        self.images = images
        self.slider.config(to=len(images))
        
        if kwargs:
            self._imshow_kwargs = kwargs

        self.select_image(1) 





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# Version number to __version__ variable
exec(open("tk_steroids/version.py").read())

install_requires = [
        'Pillow',
        'numpy',
        'matplotlib',
        ]

setuptools.setup(
    name="tk-steroids",
    version=__version__,
    author="Joni Kemppainen",
    author_email="jjtkemppainen1@sheffield.ac.uk",
    description="A collection of custom tkinter bits",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jkemppainen/tk_steroids",
    packages=setuptools.find_packages(exclude=('test')),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3) ",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.0',
)




import inspect


def inspect_booleans(function_or_method, exclude_keywords=[]):
    '''
    Inspect and get boolean default keyword arguments from a live function
    to be used in TickboxFrame.
    
    Arguments
    ---------
    function_or_method : callable
        An callable on which we apply inspect.signature
    exclude_keywords : list of strings
        Keyword arguments to be excluded by their name.

    Returns
    --------
    options, defaults : list
        Keyword argument names (string) and their default values (bool).
        See TickboxFrame for documentation.
    '''
    options = []
    defaults = []

    for name, param in inspect.signature(function_or_method).parameters.items():
        if param.kind.name in ['POSITIONAL_OR_KEYWORD', 'KEYWORD_ONLY']:
            if isinstance(param.default, bool) and name not in exclude_keywords:
                options.append(name)
                defaults.append(param.default)


    return options, defaults


def extend_keywords(keyword_source):
    '''
    Function decorator to extend signature about keyword arguments
    of a function using keywords of another function.

    Useful when you have a wrapper function with **kwargs that get
    directly passed to another function, and you want to use inspecting
    on the wrapper function and gain the keyword arguments of the
    inner function as well.

    Example usge:

        @extend_keywords(hidden_function)
        def wrapping_function(x, a=1, **kwargs):
            ...
            in_between_results = hidden_function(things, **kwargs)
            ...
            return something

        # now using inspect.signature on wrapping function contains also
        # the keywords arguments of the inner fucntion
    
    Arguments
    ---------
    keyword_source : callable
        The source function, which's kws are added to the decorated function.
    '''
    def extended(target_function):
        
        target_signature = inspect.signature(target_function)
        target_param_names = [name for name in target_signature.parameters]

        params_to_add = {}
        kwparams_to_add = {} # VAR_KEYWORD has to be the last if present

        # All parameters of the annotated function
        for name, param in target_signature.parameters.items():
            if param.kind.name not in ['VAR_KEYWORD']:
                params_to_add[name] = param
            else:
                kwparams_to_add[name] = param
        
        # Parameters of the extension source function; Only POSITIONAL_OR_KEYWORD
        # arguments are taken and converted into KEYWORD_ONLY arguments
        for name, param in inspect.signature(keyword_source).parameters.items():
            if param.kind.name in ['POSITIONAL_OR_KEYWORD'] and name not in target_param_names:
                params_to_add[name] = param.replace(kind=inspect.Parameter.KEYWORD_ONLY)

        # Make the new signature and set it as the __signature__ attribute
        # that inspection tools seem to respect
        new_signature = target_signature.replace(
                parameters=list({**params_to_add, **kwparams_to_add}.values()))
        
        target_function.__signature__ = new_signature
        
        return target_function
    return extended




import tkinter as tk
import tkinter.scrolledtext

class Listbox(tk.Frame):
    '''
    Essentially tkinter's Listbox rewrapped.
    
    At initialization, a list of selectable options are passed together with a callback
    function, which on selection is called using the current selection as the input
    argument.
    
    Attributes
    ----------
    selections
    current
    '''

    def __init__(self, parent, selections, callback, maintain_selected=True):
        '''
        SELECTIONS
        A list of strings that make up the listbox. The selection is passed
        to the callback function.
        
        CALLBACK
        Set the callback function that is called when any change or selection
        in the listbox happens. The only argument that the callback function gets
        is the selection (as shown) or None if no selection or error happens.

        The current selection is passed as the one and only argument to the callback function.
        
        maintain_selected : bool
            If true, clicking other Listboxes or widgets does not make the current
            selection to None (deselecting the selected)
        '''
        
        tk.Frame.__init__(self, parent)
        self.parent = parent
        
        self.maintain_selected = maintain_selected

        self.listbox = tk.Listbox(self, height=20)
        self.listbox.grid(sticky='NSEW')
       
        self.scrollbar= tk.Scrollbar(self, orient='vertical', command=self.listbox.yview)
        self.scrollbar.grid(row=0, column=1, sticky='NS')
        
        self.listbox.config(yscrollcommand=self.scrollbar.set)


        self.set_selections(selections)
        
        self.listbox.bind('<<ListboxSelect>>', lambda x: self._errorchecked(callback))
        
        # Make the listbox to stretch in North-South to take all the available space
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        
        self._previous_selection = None


    def _errorchecked(self, callback):
        '''
        Does some error checking related to the selection(?)        
        '''
        try:
            sel = self.listbox.curselection()[0]
            argument = self.selections[sel]
            self._previous_selection = argument
        except:
            argument = None

        if not argument is None and callable(callback):
            callback(self.selections[sel])

    def set_selections(self, selections, colors=None):
        '''
        Allows resetting the selections.

        colors          A list of valid tkinter colors, one for each selection
        '''
        
        # Empty current as it may have old entries
        self.listbox.delete(0, tk.END)
        
        self.selections = selections
        
        for i_item, item in enumerate(self.selections):
            self.listbox.insert(tk.END, item)
            
            if colors:
                self.listbox.itemconfig(i_item, bg=colors[i_item])


    def disable(self):
        self.listbox.configure(state=tk.DISABLED)


    def enable(self):
        self.listbox.configure(state=tk.NORMAL)

    
    def get_current(self):
        try:
            sel = self.listbox.curselection()[0]
            return self.selections[sel]
        except:
            return self._previous_selection

    @property
    def current(self):
        return self.get_current()


class TickboxFrame(tk.Frame):
    '''
    A series of tickboxes (Checkbuttons) and getting their True/False values.
    
    Attributes
    ----------
    states : dict
        True/False
    ticked : list
        self.states keys that have True value
    checkbuttons : list
        tk.Checkbutton objects
    '''

    def __init__(self, parent, options, fancynames=None, defaults=None, ncols=3,
            single_select=False,
            callback=None):
        '''
        parent
            Tkinter parent widget
        options : list of strings
            Names of the options
        fancynames : list of strings
            Names to show on the gui
        defaults : list of bools
            Start values, True for ticked and False for unticked
        ncols : int
            Number of columns
        single_select : bool
            If True, use Radiobuttons instead of Checkbuttons to only have
            one active selection at a time.
        callback : callable
            Executed when there's a change in the tickboxes.
            Gets the self.states dict as an input argument.
        '''
         
        tk.Frame.__init__(self, parent)
        self.parent = parent
        self._single_select = single_select 

        if single_select:
            intvar = tk.IntVar()
            self.__states = {option: intvar for option in options}
            self._option_names = options
            try:
                intvar.set(defaults.index(True))
            except:
                intvar.set(0)

        else:
            self.__states = {option: tk.IntVar() for option in options}
        
            if defaults is not None:
                for option, default in zip(options, defaults):
                    self.__states[option].set( int(default) ) 
            

        if fancynames is None:
            fancynames = {option: option for option in options}
        else:
            fancynames = {option: fancyname for option, fancyname in zip(options, fancynames)}
        
        if single_select:
            self.checkbuttons = [tk.Radiobutton(self, text=fancynames[option],
                variable=self.__states[option], command=callback,
                value=i_option) for i_option, option in enumerate(options)]
        else:
            self.checkbuttons = [tk.Checkbutton(self, text=fancynames[option],
                variable=self.__states[option], command=callback) for option in options]


        i_row = 1
        i_col = 1
        for button in self.checkbuttons:
            button.grid(row=i_row, column=i_col)
            
            i_col += 1
            if i_col > ncols:
                i_col = 1
                i_row += 1
    
    @property
    def states(self):
        if self._single_select:
            return {option: int(self.__states[option].get()) == i for i, option in enumerate(self._option_names)}
        else:
            return {option: bool(intvar.get()) for option, intvar in self.__states.items()}

    @property
    def ticked(self):
        return [s for s, b in self.states.items() if b]


class DropdownList(tk.Frame):
    '''
    A drop-in replacement for TickboxFrame using tkinter's OptionMenu
    (looks just like OptionMenu)

    Attributes
    ----------
    option_menu : object
        Underlying tkinter OptionMenu object
    label : object
        Underlying tkinter Label object (if label was specified at init)
    '''

    def __init__(self, parent, options, fancynames=None, label=None,
            default=None, callback=None, **kwargs):
        
        tk.Frame.__init__(self, parent)
        
        self.__options = options
        
   
        # Check fancynames to show
        if fancynames is None:
            fancynames = options
        else:
            if len(fancynames) != len(options):
                raise ValueError('options and fancynames different lengths, {} vs {}'.format(
                    len(fancynames), len(options)))
        
        self._fancynames = fancynames

        self._state = tk.StringVar(self)
        self._state.set(fancynames[0])
        if callback is not None:
            if callable(callback):
                self._state.trace('w', lambda *args: callback())
            else:
                raise ValueError('callback has to be callable or none, now {}'.format(callback))

        self.option_menu = tk.OptionMenu(self, self._state, *fancynames)
        self.option_menu.grid(row=1, column=2, sticky='NSWE')

        if label:
            self.label = tk.Label(self, text=label)
            self.label.grid(row=1, column=1, sticky='NSWE')
    
    @property
    def states(self):
        index = self._fancynames.index(self._state.get())
        current = self.__options[index]
        return {option: current == option for option in self.__options}

    @property
    def ticked(self):
        '''
        Returns a list of one item, the current selection.
        '''
        index = self._fancynames.index(self._state.get())
        return [self.__options[index]]



class Tabs(tk.Frame):
    '''
    Tabs widget. Can contain any tkinter widgets.

    Attributes
    ----------
    i_current : int
        Index of the currently selected tab.
    buttons : list of objects
        List of tk.Button instances.
    pages : list of objects
        A list of Tkinter widgets the tab holds.
    
    '''
    def __init__(self, parent, tab_names,
            elements=None, draw_frame=False,
            on_select_callback=None):
        '''
        Initializing the tabs.
        
        Arguments
        ---------
        parent
            Tkinter parent widget
        tab_names
            Human readable names, shown in the buttons
        elements : None or list of classes
            If None (by default), initializes tk.Frames as tabs.
            You can get these tk.Frames are in pages attribute.

            Can also classes, that get initialized as this Tabs class
            as the sole argument.
        draw_frame : bool
            If True, draw an extra frame to confine the tabs
        on_select_callback : callable
            Callback that is executed just before changing the tab.
            Has to take in one argument that is new i_current (integer).
        
        '''
        
        # Call init from LabelFrame if framing is wanted
        # Not sure if this can cause problems (when inheriting from tk.Frame)
        if draw_frame:
            tk.LabelFrame.__init__(self, parent)
        else:
            tk.Frame.__init__(self, parent)

        self.parent = parent

        self.on_select_callback = on_select_callback

        self.i_current = 0

        self.buttons = []
        self.pages = []


        buttons_frame = tk.Frame(self)
        buttons_frame.grid()

        if elements is None:
            elements = [tk.Frame for i_tab in tab_names]

        # Initialize content/elements
        for i_button, (name, element) in enumerate(zip(tab_names, elements)):

            initialized_element = element(self)
            self.pages.append(initialized_element)
            

            button = tk.Button(buttons_frame, text=name, command=lambda i_button=i_button: self.set_page(i_button))
            button.grid(row=0, column = i_button, sticky='N')
            self.buttons.append(button)
            

        self.pages[self.i_current].grid(row=1, columnspan=len(self.buttons), sticky='NSEW')
        
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)
    
    @property
    def tabs(self):
        return self.pages

    def set_page(self, i_page):
        '''
        When button number i_button is pressed.
        '''
        # Update i_current and take i_old for now
        i_old = self.i_current
        self.i_current = i_page

        if self.on_select_callback is not None:
            self.on_select_callback(self.i_current)

        # Remove the previously gridded widget
        self.pages[i_old].grid_remove()

        # Grid the new widget
        self.pages[self.i_current].grid(row=1, columnspan=len(self.buttons), sticky='NSEW')

        # Change button reliefs (pressed)
        self.buttons[i_old].config(relief=tk.RAISED)
        self.buttons[i_page].config(relief=tk.SUNKEN)


    def get_elements(self):
        '''
        Returns the initialized elements which have to the Tab as their master/parent.
        '''
        return self.pages



class ButtonsFrame(tk.Frame):
    '''
    If you just need a frame with simply buttons (with a callback) next to each other,
    use this widget.
    
    Attributes
    ----------
    buttons : list of objects
        Tkinter button objects
    label : object or None
        If label given at init, store the tk.Label object in this attibute
    '''

    def __init__(self, parent, button_names, button_commands,
            title='', label='', horizontal=True):
        '''
        Arguments
        ---------
        horizontal : bool
            If True, grid buttons horizontally. If False, grid vertically.
        title : string
            If set, init using LabelFrame that encloses the buttons and
            use this text (title) as the label.
        label : string
            Alternative or complementary to the title option, just adds a tk.Label
            as the first button (if avoiding the LabelFrame box is desired).
        '''
        tk.Frame.__init__(self, parent)
        self.parent = parent
        
        if title:
            target = tk.LabelFrame(self, text=title)
            target.grid()
        else:
            target = self

        if label:
            self.label = tk.Label(self, text=label)
            if horizontal:
                self.label.grid(row=1, column=0)
            else:
                self.label.grid(row=0, column=1)
        else:
            self.label = None

        self.buttons = []

        for i_button, (name, command) in enumerate(zip(button_names, button_commands)):
            button = tk.Button(target, text=name, command=command)
            
            if horizontal:
                button.grid(row=1, column=i_button+1)
            else:
                button.grid(row=i_button+1, column=1)

            self.buttons.append(button)


    def get_buttons(self):
        '''
        Returns the initialized buttons in the order that the buttons_kwargs
        were delivered in the ButtonsFrame constructor.
        '''
        return self.buttons



class BufferShower(tk.Frame):
    '''
    Redirect any string buffer to be printed on this buffer reader.
    Bit like a non-interactive console window.
    '''
    def __init__(self, parent, string_buffer, max_entries=100):
        '''
        string_buffer       Like StringIO, or sys.stdout
        '''
        tk.Frame.__init__(self, parent)

        self.parent = parent    
        self.string_buffer = string_buffer
        self.max_entries = max_entries
        
        self.entries = 0
        self.offset = 0

        self.text = tkinter.scrolledtext.ScrolledText(self)
        self.text.grid()
        
        self.parent.after(20, self.callback)
        
    def callback(self):
        self.string_buffer.seek(self.offset)

        for line in self.string_buffer:

            if self.entries > self.max_entries:
                self.text.delete('1.0','2.0')

            self.text.insert(tk.END, line)
            self.text.yview(tk.END)
            self.entries += 1
        
        self.offset = self.string_buffer.tell()

        self.parent.after(20, self.callback)
    

class ColorExplanation(tk.Frame):
    '''
    If colors were used in the GUI, this widget can be used easily to
    create help texts to explain meaning of the colors.
    '''

    def __init__(self, parent, colors, help_strings):
        tk.Frame.__init__(self, parent)

        for i_row, (color, string) in enumerate(zip(colors, help_strings)):
            tk.Canvas(self, width=30, height=15, bg=color).grid(row=i_row, column=0, sticky='W')
            tk.Label(self, text=string, font=('System', 8)).grid(row=i_row, column=1, sticky='W')
 


def main():
    pass

if __name__ == "__main__":
    main()
